@@ -4,26 +4,40 @@ use strict;
 
 use vars qw($VERSION);
 
-$VERSION = '0.19';
+$VERSION = '0.29';
 
 use base qw(Class::Container);
 
-use Apache::Session 1.6;
+use Apache::Session 1.81;
 
 use Exception::Class ( 'Apache::Session::Wrapper::Exception::NonExistentSessionID' =>
 		       { description => 'A non-existent session id was used',
 			 fields => [ 'session_id' ] },
                        'Apache::Session::Wrapper::Exception::Params' =>
-		       { description => 'A non-existent session id was used',
+		       { description => 'An invalid parameter or set of parameters was given',
                          alias => 'param_error' },
 		     );
 
 use Params::Validate 0.70;
-use Params::Validate qw( validate SCALAR UNDEF BOOLEAN OBJECT );
+use Params::Validate qw( validate SCALAR UNDEF BOOLEAN ARRAYREF OBJECT );
 Params::Validate::validation_options( on_fail => sub { param_error( join '', @_ ) } );
 
+use Scalar::Util ();
 
-my @HeaderMethods = qw( err_header_out  err_headers_out header_out headers_out );
+
+my $MOD_PERL = _find_mp_version();
+sub _find_mp_version
+{
+    return 0 unless $ENV{MOD_PERL};
+
+    return
+        ( $ENV{MOD_PERL} =~ /(?:1\.9|2\.\d)/
+          ? 2
+          : 1
+        );
+}
+
+my @HeaderMethods = qw( err_headers_out headers_out );
 
 my %params =
     ( always_write =>
@@ -215,6 +229,10 @@ my %params =
 	optional => 1,
 	descr => 'Path used by Apache::Session::PHP' },
 
+      session_id =>
+      { type => SCALAR,
+	optional => 1,
+	descr => 'Try this session id first when making a session' },
     );
 
 # What set of parameters are required for each session class.
@@ -238,7 +256,7 @@ my %ApacheSessionParams =
 my %OptionalApacheSessionParams =
     ( MySQL    => [ [ qw( table_name password lock_password ) ] ],
       Postgres => [ [ qw( table_name password ) ] ],
-      Informix => [ [ qw( long_read_len table_name password ) ] ],
+      Informix => [ [ qw( table_name password ) ] ],
       Oracle   => [ [ qw( long_read_len table_name password ) ] ],
       Sybase   => [ [ qw( textsize table_name password ) ] ],
     );
@@ -250,6 +268,7 @@ my %ApacheSessionFlexParams =
 	Postgres => $ApacheSessionParams{Postgres},
 	File     => [ [ qw( directory ) ] ],
 	DB_File  => [ [ qw( file_name ) ] ],
+	PHP      => $ApacheSessionParams{PHP},
       },
       lock =>
       { MySQL     => [ [ qw( lock_data_source lock_user_name ) ],
@@ -268,19 +287,58 @@ my %ApacheSessionFlexParams =
 	Base64   => [ [ ] ],
 	Sybase   => [ [ ] ],
 	UUEncode => [ [ ] ],
+	PHP      => [ [ ] ],
       },
     );
 
-__PACKAGE__->valid_params(%params);
-
 @{ $ApacheSessionFlexParams{store} }{ qw( Informix Oracle Sybase ) } =
     ( $ApacheSessionFlexParams{store}{Postgres} ) x 3;
 
 my %OptionalApacheSessionFlexParams =
-    ( Sybase => { store => [ qw( textsize ) ] },
-      Oracle => { store => [ qw( long_read_len ) ] },
+    ( store => { map { $_ => $OptionalApacheSessionParams{$_} }
+                 qw( MySQL Postgres Informix Oracle Sybase ) },
     );
 
+sub _SetValidParams {
+    my $class = shift;
+
+    my %extra;
+    for my $hash ( \%ApacheSessionParams,
+                   \%OptionalApacheSessionParams,
+                   @ApacheSessionFlexParams{ qw( store lock generate serialize ) },
+                   @OptionalApacheSessionFlexParams{ qw( store lock generate serialize ) },
+                 )
+    {
+        for my $p ( map { @$_ } map { @$_ } values %$hash )
+        {
+            $extra{$p} = { optional => 1 };
+        }
+    }
+
+    $class->valid_params( %extra, %params );
+    $class->SetStudlyForms();
+}
+__PACKAGE__->_SetValidParams();
+
+my %StudlyForm;
+sub SetStudlyForms
+{
+    %StudlyForm =
+        ( map { $_ => _studly_form($_) }
+          map { ref $_ ? @$_ :$_ }
+          map { @$_ }
+          ( values %ApacheSessionParams ),
+          ( values %OptionalApacheSessionParams ),
+          ( map { values %{ $ApacheSessionFlexParams{$_} } }
+            keys %ApacheSessionFlexParams ),
+          ( map { values %{ $OptionalApacheSessionFlexParams{$_} } }
+            keys %OptionalApacheSessionFlexParams ),
+        );
+
+    # why Apache::Session does this I do not know
+    $StudlyForm{textsize} = 'textsize';
+}
+
 sub _studly_form
 {
     my $string = shift;
@@ -288,27 +346,66 @@ sub _studly_form
     return $string;
 }
 
-my %StudlyForm =
-    ( map { $_ => _studly_form($_) }
-      map { ref $_ ? @$_ :$_ }
-      map { @$_ }
-      ( values %ApacheSessionParams ),
-      ( values %OptionalApacheSessionParams ),
-      ( map { values %{ $ApacheSessionFlexParams{$_} } }
-	keys %ApacheSessionFlexParams ),
-      ( map { values %{ $OptionalApacheSessionFlexParams{$_} } }
-	keys %OptionalApacheSessionFlexParams ),
-    );
+sub RegisterClass {
+    my $class = shift;
+    my %p = validate( @_, { name => { type => SCALAR },
+                            required => { type => SCALAR | ARRAYREF, default => [ [ ] ] },
+                            optional => { type => SCALAR | ARRAYREF, default => [ [ ] ] },
+                          },
+                    );
 
-# why Apache::Session does this I do not know
-$StudlyForm{textsize} = 'textsize';
+    $p{name} =~ s/^Apache::Session:://;
+
+    $ApacheSessionParams{ $p{name} } =
+        ( ref $p{required}
+          ? $p{required}
+          : $ApacheSessionParams{ $p{required} }
+        );
+
+    $OptionalApacheSessionParams{ $p{name} } =
+        ( ref $p{optional}
+          ? [ $p{optional} ]
+          : $OptionalApacheSessionParams{ $p{optional} }
+        );
+
+    $class->_SetValidParams();
+}
+
+sub RegisterFlexClass {
+    my $class = shift;
+    my %p = validate( @_, { type => { type => SCALAR,
+                                      regex => qr/^(?:store|lock|generate|serialize)/,
+                                    },
+                            name => { type => SCALAR },
+                            required => { type => SCALAR | ARRAYREF, default => [ [ ] ] },
+                            optional => { type => SCALAR | ARRAYREF, default => [ [ ] ] },
+                          },
+                    );
+
+    $p{name} =~ s/^Apache::Session:://;
+    $p{name} =~ s/^\Q$p{type}\E:://i;
+
+    $ApacheSessionFlexParams{ $p{type} }{ $p{name} } =
+        ( ref $p{required}
+          ? $p{required}
+          : $ApacheSessionFlexParams{ $p{type} }{ $p{required} }
+        );
+
+    $OptionalApacheSessionFlexParams{ $p{type} }{ $p{name} } =
+        ( ref $p{optional}
+          ? [ $p{optional} ]
+          : $OptionalApacheSessionFlexParams{ $p{type} }{ $p{optional} }
+        );
+
+    $class->_SetValidParams();
+}
 
 sub new
 {
-
     my $class = shift;
+    my %p = @_;
 
-    my $self = $class->SUPER::new(@_);
+    my $self = $class->SUPER::new(%p);
 
     $self->_check_session_params;
     $self->_set_session_params;
@@ -319,10 +416,14 @@ sub new
             "The header_object parameter is required in order to use cookies outside of mod_perl";
     }
 
-    eval "require Apache::Session::$self->{session_class_piece}";
-    die $@ if $@;
+    my $session_class = "Apache::Session::$self->{session_class_piece}";
+    unless ( $session_class->can('TIEHASH') )
+    {
+        eval "require $session_class";
+        die $@ if $@;
+    }
 
-    $self->_make_session;
+    $self->_make_session( $p{session_id} );
 
     $self->_bake_cookie
         if $self->{use_cookie} && ! $self->{cookie_is_baked};
@@ -364,34 +465,18 @@ sub _check_sets
     my $type = shift;
     my $class = shift;
 
-    my $matched = 0;
+    my @missing;
     foreach my $set (@$sets)
     {
-        # Don't check for missing elements unless at least one element
-        # is present.
-        if ( grep { exists $self->{$_} } @$set )
-        {
-            $matched = 1;
-        }
-        else
-        {
-            next;
-        }
+        my @matched = grep { exists $self->{$_} } @$set;
 
-        my @missing = grep { ! exists $self->{$_} } @$set;
+        return if @matched == @$set;
 
-        # At least one set has been checked if we got this far, and
-        # that's enough.
-        last unless @missing;
-
-        param_error "Some of the required parameters for your chosen $type class ($class) were missing: @missing."
-            if @missing;
+        @missing = grep { ! exists $self->{$_} } @$set;
     }
 
-    param_error "None of the required parameters for your chosen $type class ($class) were provided."
-        unless $matched;
-
-    return;
+    param_error "Some or all of the required parameters for your chosen $type class ($class) were provided."
+                . "  The following parameters were missing: @missing\n";
 }
 
 sub _set_session_params
@@ -428,24 +513,55 @@ sub _set_session_params
 
     $self->{params} = \%params;
 
-    if ( $self->{use_cookie} )
-    {
-        if ( $ENV{MOD_PERL} )
-        {
-            eval { require Apache::Cookie };
-            unless ($@)
-            {
-                $self->{cookie_class} = 'Apache::Cookie';
-                $self->{new_cookie_args} = [ Apache->request ];
-            }
-        }
+    $self->_set_cookie_fields
+        if $self->{use_cookie};
+}
 
-        unless ( $self->{cookie_class} )
-        {
-            require CGI::Cookie;
-            $self->{cookie_class} = 'CGI::Cookie';
-            $self->{new_cookie_args} = [];
-        }
+sub _set_cookie_fields
+{
+    my $self = shift;
+
+    my $cookie_class;
+    if ($MOD_PERL)
+    {
+        $cookie_class =
+            $MOD_PERL == 2 ? 'Apache2::Cookie' : 'Apache::Cookie';
+
+        eval "require $cookie_class"
+            unless $cookie_class->can('new');
+    }
+
+    unless ( $cookie_class && $cookie_class->can('new' ) )
+    {
+        require CGI::Cookie;
+        $cookie_class = 'CGI::Cookie';
+    }
+
+    $self->{cookie_class} = $cookie_class;
+
+    if ( $self->{cookie_class} eq 'CGI::Cookie' )
+    {
+        $self->{new_cookie_args} = [];
+        $self->{fetch_cookie_args} = [];
+    }
+    else
+    {
+        $self->{new_cookie_args} =
+            [ $MOD_PERL == 2
+              ? Apache2::RequestUtil->request
+              : Apache->request
+            ];
+
+        $self->{fetch_cookie_args} =
+            ( $MOD_PERL == 2
+              ? $self->{new_cookie_args}
+              : []
+            );
+        $self->{bake_cookie_args} =
+            ( $MOD_PERL == 2
+              ? $self->{new_cookie_args}
+              : []
+            );
     }
 }
 
@@ -471,15 +587,10 @@ sub _sets_to_params
 sub _make_session
 {
     my $self = shift;
-    my %p = validate( @_,
-		      { session_id =>
-			{ type => SCALAR,
-                          optional => 1,
-			},
-		      } );
+    my $session_id = shift;
 
     return if
-        defined $p{session_id} && $self->_try_session_id( $p{session_id} );
+        defined $session_id && $self->_try_session_id( $session_id );
 
     my $id = $self->_get_session_id;
     return if defined $id && $self->_try_session_id($id);
@@ -522,11 +633,19 @@ sub _get_session_id_from_cookie
 {
     my $self = shift;
 
-    my %c = $self->{cookie_class}->fetch;
-
-    return $c{ $self->{cookie_name} }->value
-        if exists $c{ $self->{cookie_name} };
+    if ( $MOD_PERL == 2 )
+    {
+        my $jar = Apache2::Cookie::Jar->new( @{ $self->{fetch_cookie_args} } );
+        my $c   = $jar->cookies( $self->{cookie_name} );
+        return $c->value if $c;
+    }
+    else
+    {
+        my %c = $self->{cookie_class}->fetch( @{ $self->{fetch_cookie_args} } );
 
+        return $c{ $self->{cookie_name} }->value
+            if exists $c{ $self->{cookie_name} };
+    }
     return undef;
 }
 
@@ -548,7 +667,7 @@ sub _try_session_id
                 $session_id, $self->{params};
 	};
 
-        if ( $@ || ! tied %s )
+        if ( $@ || ! tied %s || ! $s{_session_id} )
         {
             $self->_handle_tie_error( $@, $session_id );
             return;
@@ -571,7 +690,7 @@ sub _handle_tie_error
     my $err = shift;
     my $session_id = shift;
 
-    if ( $err =~ /Object does not exist/ )
+    if ( $err =~ /Object does not exist/ && defined $session_id )
     {
         return if $self->{allow_invalid_id};
 
@@ -581,7 +700,9 @@ sub _handle_tie_error
     }
     else
     {
-        die $@;
+        my $error =
+            $@ ? $@ : "Tying to Apache::Session::$self->{session_class_piece} failed but did not throw an exception";
+        die $error;
     }
 }
 
@@ -590,7 +711,6 @@ sub _bake_cookie
     my $self = shift;
 
     my $expires = shift || $self->{cookie_expires};
-
     $expires = undef if defined $expires && $expires =~ /^session$/i;
 
     my $domain = $self->{cookie_domain};
@@ -600,19 +720,24 @@ sub _bake_cookie
             ( @{ $self->{new_cookie_args} },
               -name    => $self->{cookie_name},
               -value   => $self->{session_id},
-              -expires => $expires,
-              ( defined $domain ?
-                ( -domain  => $domain ) :
-                ()
+              ( defined $expires
+                ? ( -expires => $expires )
+                : ()
+              ),
+              ( defined $domain
+                ? ( -domain  => $domain )
+                : ()
               ),
               -path    => $self->{cookie_path},
               -secure  => $self->{cookie_secure},
             );
 
-    if ( $cookie->can('bake') )
+    # If not running under mod_perl, CGI::Cookie->bake() will call
+    # print() to send a cookie header right now, which may not be what
+    # the user wants.
+    if ( $cookie->can('bake') && ! $cookie->isa('CGI::Cookie') )
     {
-        # Apache::Cookie
-        $cookie->bake;
+        $cookie->bake( @{ $self->{bake_cookie_args} } );
     }
     else
     {
@@ -621,7 +746,7 @@ sub _bake_cookie
         {
             if ( $header_object->can($meth) )
             {
-                $header_object->$meth( 'Set-Cookie' => $cookie );
+                $header_object->$meth->add( 'Set-Cookie' => $cookie );
                 last;
             }
         }
@@ -637,10 +762,16 @@ sub _bake_cookie
 sub session
 {
     my $self = shift;
+    my %p = validate( @_,
+		      { session_id =>
+			{ type => SCALAR,
+                          optional => 1,
+			},
+		      } );
 
-    if ( ! $self->{session} || @_ )
+    if ( ! $self->{session} || %p )
     {
-        $self->_make_session(@_);
+        $self->_make_session( $p{session_id} );
 
         $self->_bake_cookie
             if $self->{use_cookie} && ! $self->{cookie_is_baked};
@@ -733,7 +864,7 @@ This method creates a new C<Apache::Session::Wrapper> object.
 If the parameters you provide are not correct (wrong type, missing
 parameters, etc.), this method throws an
 C<Apache::Session::Wrapper::Exception::Params> exception.  You can
-treat this as a string if you want.
+treat this exception as a string if you want.
 
 =item * session
 
@@ -756,6 +887,81 @@ simply passed through to C<Apache::Session>.  For this reason, you are
 advised to familiarize yourself with the C<Apache::Session>
 documentation before attempting to configure this module.
 
+You can also register C<Apache::Session> classes, or the classes used
+for doing the work in C<Apache::Session::Flex>. See L<REGISTERING
+CLASSES> for details.
+
+=head2 Supported Classes
+
+The following classes are already supported and do not require
+registration:
+
+=over 4
+
+=item * Apache::Session::MySQL
+
+=item * Apache::Session::Postgres
+
+=item * Apache::Session::Oracle
+
+=item * Apache::Session::Informix
+
+=item * Apache::Session::Sybase
+
+=item * Apache::Session::File
+
+=item * Apache::Session::DB_File
+
+=item * Apache::Session::PHP
+
+=item * Apache::Session::Flex
+
+=back
+
+The following classes can be used with C<Apache::Session::Flex>:
+
+=over 4
+
+=item * Apache::Session::Store::MySQL
+
+=item * Apache::Session::Store::Postgres
+
+=item * Apache::Session::Store::Informix
+
+=item * Apache::Session::Store::Oracle
+
+=item * Apache::Session::Store::Sybase
+
+=item * Apache::Session::Store::File
+
+=item * Apache::Session::Store::DB_File
+
+=item * Apache::Session::Store::PHP
+
+=item * Apache::Session::Lock::MySQL
+
+=item * Apache::Session::Lock::File
+
+=item * Apache::Session::Lock::Null
+
+=item * Apache::Session::Lock::Semaphore
+
+=item * Apache::Session::Generate::MD5
+
+=item * Apache::Session::Generate::ModUsertrack
+
+=item * Apache::Session::Serialize::Storable
+
+=item * Apache::Session::Serialize::Base64
+
+=item * Apache::Session::Serialize::Sybase
+
+=item * Apache::Session::Serialize::UUEncode
+
+=item * Apache::Session::Serialize::PHP
+
+=back
+
 =head2 Generic Parameters
 
 =over 4
@@ -786,6 +992,11 @@ will be thrown instead.
 
 This defaults to true.
 
+=item * session_id  =>  string
+
+Try this session id first and use it if it exist. If the session does
+not exist, it will ignore this parameter and make a new session.
+
 =back
 
 =head2 Cookie-Related Parameters
@@ -794,11 +1005,9 @@ This defaults to true.
 
 =item * use_cookie  =>  boolean
 
-If true, then this module will use C<Apache::Cookie> or C<CGI::Cookie>
-(as appropriate) to set and read cookies that contain the session id.
-
-The cookie will be set again every time the client accesses a Mason
-component unless the C<cookie_resend> parameter is false.
+If true, then this module will use one of C<Apache::Cookie>,
+C<Apache2::Cookie> or C<CGI::Cookie> (as appropriate) to set and read
+cookies that contain the session id.
 
 =item * cookie_name  =>  name
 
@@ -840,13 +1049,12 @@ cookie has the effect of updating the expiration time.
 =item * header_object => object
 
 When running outside of mod_perl, you must provide an object to which
-the cookie header can be added.  This object must provide one of the
-following methods: C<err_header_out()>, C<err_headers_out()>,
-C<header_out()>, or C<headers_out()>.  The pluralized versions are
-needed to support mod_perl 2.
+the cookie header can be added.  This object must provide an
+C<err_headers_out()> or C<headers_out()> method.
 
-Under mod_perl, this will default to the object returned by C<<
-Apache->request >>.  This may not work under mod_perl 2.
+Under mod_perl 1, this will default to the object returned by C<<
+Apache->request() >>. Under mod_perl 2 we call C<<
+Apache2::RequestUtil->request() >>
 
 =back
 
@@ -863,8 +1071,6 @@ name of the parameter that is checked.
 If you are also using cookies, then the module checks the param object
 I<first>, and then it checks for a cookie.
 
-The session id is available from C<< $m->session->{_session_id} >>.
-
 =item * param_object  =>  object
 
 This should be an object that provides a C<param()> method.  This
@@ -1010,13 +1216,109 @@ When run under mod_perl, this module attempts to first use
 C<Apache::Cookie> for cookie-handling.  Otherwise it uses
 C<CGI::Cookie> as a fallback.
 
-If it ends up using C<CGI::Cookie> then must provide a "header_object"
-parameter.  The module calls C<err_header_out()> or C<header_out()> on
-the provided object, using the former if it's available.
+If it ends up using C<CGI::Cookie> then you must provide a
+"header_object" parameter. This object must have an
+C<err_headers_out()> or C<headers_out()> method. It looks for these
+methods in that order. The method is expected to return an object with
+an API like C<Apache::Table>. It calls C<add()> on the returned method
+to add a "Set-Cookie" header.
+
+=head1 REGISTERING CLASSES
+
+In order to support any C<Apache::Session> subclasses, this module
+provides a simple registration mechanism.
+
+You can register an C<Apache::Session> subclass, or a class intended
+to provide a class that implements something required by
+C<Apache::Session::Flex>.
+
+=head2 Registering a Complete Subclass
+
+This is done by calling C<< Apache::Session::Wrapper->RegisterClass() >>:
+
+  Apache::Session::Wrapper->RegisterClass
+      ( name     => 'MyClass',
+        required => [ [ qw( param1 param2 ) ],
+                      [ qw( param3 param4 ) ] ],
+        optional => [ 'optional_p' ],
+      );
+
+  Apache::Session::Wrapper->RegisterClass
+      ( name     => 'Apache::Session::MyFile',
+        required => 'File',
+        optional => 'File',
+      );
+
+The C<RegisterClass()> method takes the following options:
+
+=over 4
+
+=item * name
+
+This should be the name of the class you are registering. The actual
+class must start with "Apache::Session::", but this part does not need
+to be included when registering the class (it's optional).
+
+=item * required
+
+These are the required parameters for this class.
+
+The value of this parameter can either be a string or a reference to
+an array of array references.
+
+If it is a string, then it identifies an existing C<Apache::Session>
+subclass which is already registered or built-in, like "File" or
+"Postgres".
+
+If it an array reference, then I<that reference> should in turn
+contain one or more array references. Each of those contained
+references represents one set of required parameters. When an
+C<Apache::Session::Wrapper> object is constructed, only one of these
+sets must be passed in. For example:
+
+  required => [ [ qw( p1 p2 ) ],
+                [ qw( p2 p3 p4 ) ] ]
+
+This says that either "p1" and "p2" must be provided, I<or> "p2",
+"p3", and "p4".
+
+If there are no required parameters for this class, then the
+"required" parameter can be omitted.
+
+=item * optional
+
+This specifies optional parameters, and should just be a simple array
+reference.
+
+=back
+
+=head2 Registering a Subclass for Flex
+
+Registering a subclass that can be used with C<Apache::Session::Flex>
+is very similar to registering a complete class:
+
+  Apache::Session::Wrapper->RegisterFlexClass
+      ( name     => 'MyClass',
+        type     => 'Store',
+        required => [ [ qw( param1 param2 ) ],
+                      [ qw( param3 param4 ) ] ],
+        optional => [ 'optional_p' ],
+      );
+
+  Apache::Session::Wrapper->RegisterFlexClass
+      ( name     => 'Apache::Session::Store::MyFile',
+        type     => 'store',
+        required => 'File',
+        optional => 'File',
+      );
+
+The C<RegisterFlexClass()> method has the same parameters as
+C<RegisterClass()>, but it also requires a "type" parameter. This must
+be one of "store", "lock", "generate", or "serialize".
 
 =head1 SUBCLASSING
 
-This class provides a simply hook for subclasses.  Before trying to
+This class provides a simple hook for subclasses.  Before trying to
 get a session id from the URL or cookie, it calls a method named
 C<_get_session_id()>.  In this class, that method is a no-op, but you
 can override this in a subclass.
